@@ -1,21 +1,23 @@
-use super::{deques::Deques, KeyDate, KeyHashDate, ValueEntry};
+use super::{deques::Deques, AccessTime, CacheBuilder, KeyDate, KeyHashDate, ValueEntry, Weigher};
 use crate::common::{
+    self,
     deque::{CacheRegion, DeqNode, Deque},
     frequency_sketch::FrequencySketch,
     time::{CheckedTimeOps, Clock, Instant},
-    AccessTime,
 };
 
+use smallvec::SmallVec;
 use std::{
     borrow::Borrow,
     collections::{hash_map::RandomState, HashMap},
-    convert::TryInto,
     hash::{BuildHasher, Hash, Hasher},
     ptr::NonNull,
     rc::Rc,
     time::Duration,
 };
 
+const EVICTION_BATCH_SIZE: usize = 100;
+
 type CacheStore<K, V, S> = std::collections::HashMap<Rc<K>, ValueEntry<K, V>, S>;
 
 /// An in-memory cache that is _not_ thread-safe.
@@ -75,7 +77,52 @@ type CacheStore<K, V, S> = std::collections::HashMap<Rc<K>, ValueEntry<K, V>, S>
 /// }
 /// ```
 ///
-/// # Expiration Policies
+/// # Size-based Eviction
+///
+/// ```rust
+/// use std::convert::TryInto;
+/// use moka::unsync::Cache;
+///
+/// // Evict based on the number of entries in the cache.
+/// let mut cache = Cache::builder()
+///     // Up to 10,000 entries.
+///     .max_capacity(10_000)
+///     // Create the cache.
+///     .build();
+/// cache.insert(1, "one".to_string());
+///
+/// // Evict based on the byte length of strings in the cache.
+/// let mut cache = Cache::builder()
+///     // A weigher closure takes &K and &V and returns a u32
+///     // representing the relative size of the entry.
+///     .weigher(|_key, value: &String| -> u32 {
+///         value.len().try_into().unwrap_or(u32::MAX)
+///     })
+///     // This cache will hold up to 32MiB of values.
+///     .max_capacity(32 * 1024 * 1024)
+///     .build();
+/// cache.insert(2, "two".to_string());
+/// ```
+///
+/// If your cache should not grow beyond a certain size, use the `max_capacity`
+/// method of the [`CacheBuilder`][builder-struct] to set the upper bound. The cache
+/// will try to evict entries that have not been used recently or very often.
+///
+/// At the cache creation time, a weigher closure can be set by the `weigher` method
+/// of the `CacheBuilder`. A weigher closure takes `&K` and `&V` as the arguments and
+/// returns a `u32` representing the relative size of the entry:
+///
+/// - If the `weigher` is _not_ set, the cache will treat each entry has the same
+///   size of `1`. This means the cache will be bounded by the number of entries.
+/// - If the `weigher` is set, the cache will call the weigher to calculate the
+///   weighted size (relative size) on an entry. This means the cache will be bounded
+///   by the total weighted size of entries.
+///
+/// Note that weighted sizes are not used when making eviction selections.
+///
+/// [builder-struct]: ./struct.CacheBuilder.html
+///
+/// # Time-based Expirations
 ///
 /// `Cache` supports the following expiration policies:
 ///
@@ -109,9 +156,12 @@ type CacheStore<K, V, S> = std::collections::HashMap<Rc<K>, ValueEntry<K, V>, S>
 /// [ahash-crate]: https://crates.io/crates/ahash
 ///
 pub struct Cache<K, V, S = RandomState> {
-    max_capacity: usize,
+    max_capacity: Option<u64>,
+    entry_count: u64,
+    weighted_size: u64,
     cache: CacheStore<K, V, S>,
     build_hasher: S,
+    weigher: Option<Weigher<K, V>>,
     deques: Deques<K>,
     frequency_sketch: FrequencySketch,
     time_to_live: Option<Duration>,
@@ -129,9 +179,17 @@ where
     /// `time_to_live`, use the [`CacheBuilder`][builder-struct].
     ///
     /// [builder-struct]: ./struct.CacheBuilder.html
-    pub fn new(max_capacity: usize) -> Self {
+    pub fn new(max_capacity: u64) -> Self {
         let build_hasher = RandomState::default();
-        Self::with_everything(max_capacity, None, build_hasher, None, None)
+        Self::with_everything(Some(max_capacity), None, build_hasher, None, None, None)
+    }
+
+    /// Returns a [`CacheBuilder`][builder-struct], which can builds a `Cache` with
+    /// various configuration knobs.
+    ///
+    /// [builder-struct]: ./struct.CacheBuilder.html
+    pub fn builder() -> CacheBuilder<K, V, Cache<K, V, RandomState>> {
+        CacheBuilder::default()
     }
 }
 
@@ -144,9 +202,10 @@ where
     S: BuildHasher + Clone,
 {
     pub(crate) fn with_everything(
-        max_capacity: usize,
+        max_capacity: Option<u64>,
         initial_capacity: Option<usize>,
         build_hasher: S,
+        weigher: Option<Weigher<K, V>>,
         time_to_live: Option<Duration>,
         time_to_idle: Option<Duration>,
     ) -> Self {
@@ -155,18 +214,15 @@ where
             build_hasher.clone(),
         );
 
-        // Ensure skt_capacity fits in a range of `128u32..=u32::MAX`.
-        let skt_capacity = max_capacity
-            .try_into() // Convert to u32.
-            .unwrap_or(u32::MAX)
-            .max(128);
-        let frequency_sketch = FrequencySketch::with_capacity(skt_capacity);
         Self {
-            max_capacity,
+            max_capacity: max_capacity.map(|n| n as u64),
+            entry_count: 0,
+            weighted_size: 0,
             cache,
             build_hasher,
-            deques: Deques::default(),
-            frequency_sketch,
+            weigher,
+            deques: Default::default(),
+            frequency_sketch: Default::default(),
             time_to_live,
             time_to_idle,
             expiration_clock: None,
@@ -184,7 +240,8 @@ where
         Rc<K>: Borrow<Q>,
         Q: Hash + Eq + ?Sized,
     {
-        let timestamp = self.evict_if_needed();
+        let timestamp = self.evict_expired_if_needed();
+        self.evict_lru_entries();
         self.frequency_sketch.increment(self.hash(key));
 
         match (self.cache.get_mut(key), timestamp, &mut self.deques) {
@@ -213,15 +270,17 @@ where
     ///
     /// If the cache has this key present, the value is updated.
     pub fn insert(&mut self, key: K, value: V) {
-        let timestamp = self.evict_if_needed();
+        let timestamp = self.evict_expired_if_needed();
+        self.evict_lru_entries();
+        let policy_weight = weigh(&mut self.weigher, &key, &value);
         let key = Rc::new(key);
-        let entry = ValueEntry::new(value);
+        let entry = ValueEntry::new(value, policy_weight);
 
         if let Some(old_entry) = self.cache.insert(Rc::clone(&key), entry) {
-            self.handle_update(key, timestamp, old_entry);
+            self.handle_update(key, timestamp, policy_weight, old_entry);
         } else {
             let hash = self.hash(&key);
-            self.handle_insert(key, hash, timestamp);
+            self.handle_insert(key, hash, policy_weight, timestamp);
         }
     }
 
@@ -234,11 +293,14 @@ where
         Rc<K>: Borrow<Q>,
         Q: Hash + Eq + ?Sized,
     {
-        self.evict_if_needed();
+        self.evict_expired_if_needed();
+        self.evict_lru_entries();
 
         if let Some(mut entry) = self.cache.remove(key) {
+            let weight = entry.policy_weight();
             self.deques.unlink_ao(&mut entry);
-            Deques::unlink_wo(&mut self.deques.write_order, &mut entry)
+            Deques::unlink_wo(&mut self.deques.write_order, &mut entry);
+            self.saturating_sub_from_total_weight(weight as u64);
         }
     }
 
@@ -250,6 +312,7 @@ where
     pub fn invalidate_all(&mut self) {
         self.cache.clear();
         self.deques.clear();
+        self.weighted_size = 0;
     }
 
     /// Discards cached values that satisfy a predicate.
@@ -280,17 +343,22 @@ where
             .map(|(key, _)| Rc::clone(key))
             .collect::<Vec<_>>();
 
+        let mut invalidated = 0u64;
+
         keys_to_invalidate.into_iter().for_each(|k| {
             if let Some(mut entry) = cache.remove(&k) {
+                let weight = entry.policy_weight();
                 deques.unlink_ao(&mut entry);
                 Deques::unlink_wo(&mut deques.write_order, &mut entry);
+                invalidated = invalidated.saturating_sub(weight as u64);
             }
         });
+        self.saturating_sub_from_total_weight(invalidated);
     }
 
     /// Returns the `max_capacity` of this cache.
-    pub fn max_capacity(&self) -> usize {
-        self.max_capacity
+    pub fn max_capacity(&self) -> Option<usize> {
+        self.max_capacity.map(|n| n as usize)
     }
 
     /// Returns the `time_to_live` of this cache.
@@ -329,10 +397,10 @@ where
     }
 
     #[inline]
-    fn evict_if_needed(&mut self) -> Option<Instant> {
+    fn evict_expired_if_needed(&mut self) -> Option<Instant> {
         if self.has_expiry() {
             let ts = self.current_time_from_expiration_clock();
-            self.evict(ts);
+            self.evict_expired(ts);
             Some(ts)
         } else {
             None
@@ -387,9 +455,59 @@ where
         deques.move_to_back_ao(entry)
     }
 
+    fn has_enough_capacity(&self, candidate_weight: u32, ws: u64) -> bool {
+        self.max_capacity
+            .map(|limit| ws + candidate_weight as u64 <= limit)
+            .unwrap_or(true)
+    }
+
+    fn weights_to_evict(&self) -> u64 {
+        self.max_capacity
+            .map(|limit| self.weighted_size.saturating_sub(limit))
+            .unwrap_or_default()
+    }
+
     #[inline]
-    fn handle_insert(&mut self, key: Rc<K>, hash: u64, timestamp: Option<Instant>) {
-        let has_free_space = self.cache.len() <= self.max_capacity;
+    fn should_enable_frequency_sketch(&self) -> bool {
+        if let Some(max_cap) = self.max_capacity {
+            self.weighted_size >= max_cap / 2
+        } else {
+            false
+        }
+    }
+
+    #[inline]
+    fn enable_frequency_sketch(&mut self) {
+        if let Some(max_cap) = self.max_capacity {
+            let num_entries = if self.weigher.is_some() {
+                self.entry_count
+            } else {
+                max_cap
+            };
+            let skt_capacity = common::sketch_capacity(num_entries);
+            self.frequency_sketch.ensure_capacity(skt_capacity);
+        }
+    }
+
+    fn saturating_add_to_total_weight(&mut self, weight: u64) {
+        let total = &mut self.weighted_size;
+        *total = total.saturating_add(weight as u64);
+    }
+
+    fn saturating_sub_from_total_weight(&mut self, weight: u64) {
+        let total = &mut self.weighted_size;
+        *total = total.saturating_sub(weight as u64);
+    }
+
+    #[inline]
+    fn handle_insert(
+        &mut self,
+        key: Rc<K>,
+        hash: u64,
+        policy_weight: u32,
+        timestamp: Option<Instant>,
+    ) {
+        let has_free_space = self.has_enough_capacity(policy_weight, self.weighted_size);
         let (cache, deqs, freq) = (&mut self.cache, &mut self.deques, &self.frequency_sketch);
 
         if has_free_space {
@@ -404,26 +522,46 @@ where
             if self.time_to_live.is_some() {
                 deqs.push_back_wo(KeyDate::new(key, timestamp), entry);
             }
-        } else {
-            let victim = Self::find_cache_victim(deqs, freq);
-            if Self::admit(hash, victim, freq) {
-                // Remove the victim from the cache and deque.
-                //
-                // TODO: Check if the selected victim was actually removed. If not,
-                // maybe we should find another victim. This can happen because it
-                // could have been already removed from the cache but the removal
-                // from the deque is still on the write operations queue and is not
-                // yet executed.
-                if let Some(mut vic_entry) = cache.remove(&victim.element.key) {
+            self.entry_count += 1;
+            self.saturating_add_to_total_weight(policy_weight as u64);
+
+            if self.should_enable_frequency_sketch() {
+                self.enable_frequency_sketch();
+            }
+
+            return;
+        }
+
+        if let Some(max) = self.max_capacity {
+            if policy_weight as u64 > max {
+                // The candidate is too big to fit in the cache. Reject it.
+                cache.remove(&Rc::clone(&key));
+                return;
+            }
+        }
+
+        let mut candidate = EntrySizeAndFrequency::new(policy_weight as u64);
+        candidate.add_frequency(freq, hash);
+
+        match Self::admit(&candidate, cache, deqs, freq, &mut self.weigher) {
+            AdmissionResult::Admitted {
+                victim_nodes,
+                victims_weight,
+            } => {
+                // Remove the victims from the cache (hash map) and deque.
+                for victim in victim_nodes {
+                    // Remove the victim from the hash map.
+                    let mut vic_entry = cache
+                        .remove(unsafe { &victim.as_ref().element.key })
+                        .expect("Cannot remove a victim from the hash map");
+                    // And then remove the victim from the deques.
                     deqs.unlink_ao(&mut vic_entry);
                     Deques::unlink_wo(&mut deqs.write_order, &mut vic_entry);
-                } else {
-                    let victim = NonNull::from(victim);
-                    deqs.unlink_node_ao(victim);
+                    self.entry_count -= 1;
                 }
+
                 // Add the candidate to the deque.
                 let entry = cache.get_mut(&key).unwrap();
-
                 let key = Rc::clone(&key);
                 deqs.push_back_ao(
                     CacheRegion::MainProbation,
@@ -433,56 +571,120 @@ where
                 if self.time_to_live.is_some() {
                     deqs.push_back_wo(KeyDate::new(key, timestamp), entry);
                 }
-            } else {
+
+                self.entry_count += 1;
+                Self::saturating_sub_from_total_weight(self, victims_weight);
+                Self::saturating_add_to_total_weight(self, policy_weight as u64);
+
+                if self.should_enable_frequency_sketch() {
+                    self.enable_frequency_sketch();
+                }
+            }
+            AdmissionResult::Rejected => {
                 // Remove the candidate from the cache.
                 cache.remove(&key);
             }
         }
     }
 
-    #[inline]
-    fn find_cache_victim<'a>(
-        deqs: &'a mut Deques<K>,
-        _freq: &FrequencySketch,
-    ) -> &'a DeqNode<KeyHashDate<K>> {
-        // TODO: Check its frequency. If it is not very low, maybe we should
-        // check frequencies of next few others and pick from them.
-        deqs.probation.peek_front().expect("No victim found")
-    }
-
+    /// Performs size-aware admission explained in the paper:
+    /// [Lightweight Robust Size Aware Cache Management][size-aware-cache-paper]
+    /// by Gil Einziger, Ohad Eytan, Roy Friedman, Ben Manes.
+    ///
+    /// [size-aware-cache-paper]: https://arxiv.org/abs/2105.08770
+    ///
+    /// There are some modifications in this implementation:
+    /// - To admit to the main space, candidate's frequency must be higher than
+    ///   the aggregated frequencies of the potential victims. (In the paper,
+    ///   `>=` operator is used rather than `>`)  The `>` operator will do a better
+    ///   job to prevent the main space from polluting.
+    /// - When a candidate is rejected, the potential victims will stay at the LRU
+    ///   position of the probation access-order queue. (In the paper, they will be
+    ///   promoted (to the MRU position?) to force the eviction policy to select a
+    ///   different set of victims for the next candidate). We may implement the
+    ///   paper's behavior later?
+    ///
     #[inline]
     fn admit(
-        candidate_hash: u64,
-        victim: &DeqNode<KeyHashDate<K>>,
+        candidate: &EntrySizeAndFrequency,
+        cache: &CacheStore<K, V, S>,
+        deqs: &Deques<K>,
         freq: &FrequencySketch,
-    ) -> bool {
+        weigher: &mut Option<Weigher<K, V>>,
+    ) -> AdmissionResult<K> {
+        let mut victims = EntrySizeAndFrequency::default();
+        let mut victim_nodes = SmallVec::default();
+
+        // Get first potential victim at the LRU position.
+        let mut next_victim = deqs.probation.peek_front();
+
+        // Aggregate potential victims.
+        while victims.weight < candidate.weight {
+            if candidate.freq < victims.freq {
+                break;
+            }
+            if let Some(victim) = next_victim.take() {
+                next_victim = victim.next_node();
+
+                let vic_entry = cache
+                    .get(&victim.element.key)
+                    .expect("Cannot get an victim entry");
+                victims.add_policy_weight(victim.element.key.as_ref(), &vic_entry.value, weigher);
+                victims.add_frequency(freq, victim.element.hash);
+                victim_nodes.push(NonNull::from(victim));
+            } else {
+                // No more potential victims.
+                break;
+            }
+        }
+
+        // Admit or reject the candidate.
+
         // TODO: Implement some randomness to mitigate hash DoS attack.
         // See Caffeine's implementation.
-        freq.frequency(candidate_hash) > freq.frequency(victim.element.hash)
+
+        if victims.weight >= candidate.weight && candidate.freq > victims.freq {
+            AdmissionResult::Admitted {
+                victim_nodes,
+                victims_weight: victims.weight,
+            }
+        } else {
+            AdmissionResult::Rejected
+        }
     }
 
     fn handle_update(
         &mut self,
         key: Rc<K>,
         timestamp: Option<Instant>,
+        policy_weight: u32,
         old_entry: ValueEntry<K, V>,
     ) {
+        let old_policy_weight = old_entry.policy_weight();
+
         let entry = self.cache.get_mut(&key).unwrap();
         entry.replace_deq_nodes_with(old_entry);
         if let Some(ts) = timestamp {
             entry.set_last_accessed(ts);
             entry.set_last_modified(ts);
         }
+        entry.set_policy_weight(policy_weight);
+
         let deqs = &mut self.deques;
         deqs.move_to_back_ao(entry);
-        deqs.move_to_back_wo(entry)
+        if self.time_to_live.is_some() {
+            deqs.move_to_back_wo(entry);
+        }
+
+        self.saturating_sub_from_total_weight(old_policy_weight as u64);
+        self.saturating_add_to_total_weight(policy_weight as u64);
     }
 
-    fn evict(&mut self, now: Instant) {
-        const EVICTION_BATCH_SIZE: usize = 100;
-
+    fn evict_expired(&mut self, now: Instant) {
         if self.time_to_live.is_some() {
-            self.remove_expired_wo(EVICTION_BATCH_SIZE, now);
+            let (count, weight) = self.remove_expired_wo(EVICTION_BATCH_SIZE, now);
+            self.entry_count -= count;
+            self.saturating_sub_from_total_weight(weight);
         }
 
         if self.time_to_idle.is_some() {
@@ -508,12 +710,18 @@ where
                 )
             };
 
-            rm_expired_ao("window", window);
-            rm_expired_ao("probation", probation);
-            rm_expired_ao("protected", protected);
+            let (count1, weight1) = rm_expired_ao("window", window);
+            let (count2, weight2) = rm_expired_ao("probation", probation);
+            let (count3, weight3) = rm_expired_ao("protected", protected);
+
+            self.entry_count -= count1 + count2 + count3;
+            self.saturating_sub_from_total_weight(weight1);
+            self.saturating_sub_from_total_weight(weight2);
+            self.saturating_sub_from_total_weight(weight3);
         }
     }
 
+    // Returns (u64, u64) where (evicted_entry_count, evicted_policy_weight).
     #[inline]
     fn remove_expired_ao(
         deq_name: &str,
@@ -523,7 +731,10 @@ where
         time_to_idle: &Option<Duration>,
         batch_size: usize,
         now: Instant,
-    ) {
+    ) -> (u64, u64) {
+        let mut evicted_entry_count = 0u64;
+        let mut evicted_policy_weight = 0u64;
+
         for _ in 0..batch_size {
             let key = deq
                 .peek_front()
@@ -540,18 +751,29 @@ where
                 break;
             }
 
-            if let Some(mut entry) = cache.remove(&key.unwrap()) {
+            let key = key.unwrap();
+
+            if let Some(mut entry) = cache.remove(&key) {
+                let weight = entry.policy_weight();
                 Deques::unlink_ao_from_deque(deq_name, deq, &mut entry);
                 Deques::unlink_wo(write_order_deq, &mut entry);
+                evicted_entry_count += 1;
+                evicted_policy_weight = evicted_policy_weight.saturating_add(weight as u64);
             } else {
                 deq.pop_front();
             }
         }
+
+        (evicted_entry_count, evicted_policy_weight)
     }
 
+    // Returns (u64, u64) where (evicted_entry_count, evicted_policy_weight).
     #[inline]
-    fn remove_expired_wo(&mut self, batch_size: usize, now: Instant) {
+    fn remove_expired_wo(&mut self, batch_size: usize, now: Instant) -> (u64, u64) {
+        let mut evicted_entry_count = 0u64;
+        let mut evicted_policy_weight = 0u64;
         let time_to_live = &self.time_to_live;
+
         for _ in 0..batch_size {
             let key = self
                 .deques
@@ -570,13 +792,63 @@ where
                 break;
             }
 
-            if let Some(mut entry) = self.cache.remove(&key.unwrap()) {
+            let key = key.unwrap();
+
+            if let Some(mut entry) = self.cache.remove(&key) {
+                let weight = entry.policy_weight();
                 self.deques.unlink_ao(&mut entry);
                 Deques::unlink_wo(&mut self.deques.write_order, &mut entry);
+                evicted_entry_count += 1;
+                evicted_policy_weight = evicted_policy_weight.saturating_sub(weight as u64);
             } else {
                 self.deques.write_order.pop_front();
             }
         }
+
+        (evicted_entry_count, evicted_policy_weight)
+    }
+
+    #[inline]
+    fn evict_lru_entries(&mut self) {
+        const DEQ_NAME: &str = "probation";
+
+        let weights_to_evict = self.weights_to_evict();
+        let mut evicted_count = 0u64;
+        let mut evicted_policy_weight = 0u64;
+
+        {
+            let deqs = &mut self.deques;
+            let (probation, wo, cache) =
+                (&mut deqs.probation, &mut deqs.write_order, &mut self.cache);
+
+            for _ in 0..EVICTION_BATCH_SIZE {
+                if evicted_policy_weight >= weights_to_evict {
+                    break;
+                }
+
+                let key = probation
+                    .peek_front()
+                    .map(|node| Rc::clone(&node.element.key));
+
+                if key.is_none() {
+                    break;
+                }
+                let key = key.unwrap();
+
+                if let Some(mut entry) = cache.remove(&key) {
+                    let weight = entry.policy_weight();
+                    Deques::unlink_ao_from_deque(DEQ_NAME, probation, &mut entry);
+                    Deques::unlink_wo(wo, &mut entry);
+                    evicted_count += 1;
+                    evicted_policy_weight = evicted_policy_weight.saturating_add(weight as u64);
+                } else {
+                    probation.pop_front();
+                }
+            }
+        }
+
+        self.entry_count -= evicted_count;
+        self.saturating_sub_from_total_weight(evicted_policy_weight);
     }
 }
 
@@ -594,6 +866,48 @@ where
     }
 }
 
+#[derive(Default)]
+struct EntrySizeAndFrequency {
+    weight: u64,
+    freq: u32,
+}
+
+impl EntrySizeAndFrequency {
+    fn new(policy_weight: u64) -> Self {
+        Self {
+            weight: policy_weight,
+            ..Default::default()
+        }
+    }
+
+    fn add_policy_weight<K, V>(&mut self, key: &K, value: &V, weigher: &mut Option<Weigher<K, V>>) {
+        self.weight += weigh(weigher, key, value) as u64;
+    }
+
+    fn add_frequency(&mut self, freq: &FrequencySketch, hash: u64) {
+        self.freq += freq.frequency(hash) as u32;
+    }
+}
+
+// Access-Order Queue Node
+type AoqNode<K> = NonNull<DeqNode<KeyHashDate<K>>>;
+
+enum AdmissionResult<K> {
+    Admitted {
+        victim_nodes: SmallVec<[AoqNode<K>; 8]>,
+        victims_weight: u64,
+    },
+    Rejected,
+}
+
+//
+// private free-standing functions
+//
+#[inline]
+fn weigh<K, V>(weigher: &mut Option<Weigher<K, V>>, key: &K, value: &V) -> u32 {
+    weigher.as_mut().map(|w| w(key, value)).unwrap_or(1)
+}
+
 // To see the debug prints, run test as `cargo test -- --nocapture`
 #[cfg(test)]
 mod tests {
@@ -605,6 +919,7 @@ mod tests {
     #[test]
     fn basic_single_thread() {
         let mut cache = Cache::new(3);
+        cache.enable_frequency_sketch();
 
         cache.insert("a", "alice");
         cache.insert("b", "bob");
@@ -628,7 +943,7 @@ mod tests {
         assert_eq!(cache.get(&"d"), None); //   d -> 2
 
         // "d" should be admitted and "c" should be evicted
-        // because d's frequency is higher then c's.
+        // because d's frequency is higher than c's.
         cache.insert("d", "dennis");
         assert_eq!(cache.get(&"a"), Some(&"alice"));
         assert_eq!(cache.get(&"b"), Some(&"bob"));
@@ -639,9 +954,77 @@ mod tests {
         assert_eq!(cache.get(&"b"), None);
     }
 
+    #[test]
+    fn size_aware_eviction() {
+        let weigher = |_k: &&str, v: &(&str, u32)| v.1;
+
+        let alice = ("alice", 10);
+        let bob = ("bob", 15);
+        let bill = ("bill", 20);
+        let cindy = ("cindy", 5);
+        let david = ("david", 15);
+        let dennis = ("dennis", 15);
+
+        let mut cache = Cache::builder().max_capacity(31).weigher(weigher).build();
+        cache.enable_frequency_sketch();
+
+        cache.insert("a", alice);
+        cache.insert("b", bob);
+        assert_eq!(cache.get(&"a"), Some(&alice));
+        assert_eq!(cache.get(&"b"), Some(&bob));
+        // order (LRU -> MRU) and counts: a -> 1, b -> 1
+
+        cache.insert("c", cindy);
+        assert_eq!(cache.get(&"c"), Some(&cindy));
+        // order and counts: a -> 1, b -> 1, c -> 1
+
+        assert_eq!(cache.get(&"a"), Some(&alice));
+        assert_eq!(cache.get(&"b"), Some(&bob));
+        // order and counts: c -> 1, a -> 2, b -> 2
+
+        // To enter "d" (weight: 15), it needs to evict "c" (w: 5) and "a" (w: 10).
+        // "d" must have higher count than 3, which is the aggregated count
+        // of "a" and "c".
+        cache.insert("d", david); //   count: d -> 0
+        assert_eq!(cache.get(&"d"), None); //   d -> 1
+
+        cache.insert("d", david);
+        assert_eq!(cache.get(&"d"), None); //   d -> 2
+
+        cache.insert("d", david);
+        assert_eq!(cache.get(&"d"), None); //   d -> 3
+
+        cache.insert("d", david);
+        assert_eq!(cache.get(&"d"), None); //   d -> 4
+
+        // Finally "d" should be admitted by evicting "c" and "a".
+        cache.insert("d", dennis);
+        assert_eq!(cache.get(&"a"), None);
+        assert_eq!(cache.get(&"b"), Some(&bob));
+        assert_eq!(cache.get(&"c"), None);
+        assert_eq!(cache.get(&"d"), Some(&dennis));
+
+        // Update "b" with "bill" (w: 15 -> 20). This should evict "d" (w: 15).
+        cache.insert("b", bill);
+        assert_eq!(cache.get(&"b"), Some(&bill));
+        assert_eq!(cache.get(&"d"), None);
+
+        // Re-add "a" (w: 10) and update "b" with "bob" (w: 20 -> 15).
+        cache.insert("a", alice);
+        cache.insert("b", bob);
+        assert_eq!(cache.get(&"a"), Some(&alice));
+        assert_eq!(cache.get(&"b"), Some(&bob));
+        assert_eq!(cache.get(&"d"), None);
+
+        // Verify the sizes.
+        assert_eq!(cache.entry_count, 2);
+        assert_eq!(cache.weighted_size, 25);
+    }
+
     #[test]
     fn invalidate_all() {
         let mut cache = Cache::new(100);
+        cache.enable_frequency_sketch();
 
         cache.insert("a", "alice");
         cache.insert("b", "bob");
@@ -665,6 +1048,7 @@ mod tests {
         use std::collections::HashSet;
 
         let mut cache = Cache::new(100);
+        cache.enable_frequency_sketch();
 
         let (clock, mock) = Clock::mock();
         cache.set_expiration_clock(Some(clock));
@@ -708,6 +1092,7 @@ mod tests {
         let mut cache = CacheBuilder::new(100)
             .time_to_live(Duration::from_secs(10))
             .build();
+        cache.enable_frequency_sketch();
 
         let (clock, mock) = Clock::mock();
         cache.set_expiration_clock(Some(clock));
@@ -751,6 +1136,7 @@ mod tests {
         let mut cache = CacheBuilder::new(100)
             .time_to_idle(Duration::from_secs(10))
             .build();
+        cache.enable_frequency_sketch();
 
         let (clock, mock) = Clock::mock();
         cache.set_expiration_clock(Some(clock));
@@ -784,11 +1170,12 @@ mod tests {
     #[test]
     fn test_skt_capacity_will_not_overflow() {
         // power of two
-        let pot = |exp| 2_usize.pow(exp);
+        let pot = |exp| 2u64.pow(exp);
 
         let ensure_sketch_len = |max_capacity, len, name| {
-            let cache = Cache::<u8, u8>::new(max_capacity);
-            assert_eq!(cache.frequency_sketch.table_len(), len, "{}", name);
+            let mut cache = Cache::<u8, u8>::new(max_capacity);
+            cache.enable_frequency_sketch();
+            assert_eq!(cache.frequency_sketch.table_len(), len as usize, "{}", name);
         };
 
         if cfg!(target_pointer_width = "32") {
@@ -803,7 +1190,7 @@ mod tests {
             ensure_sketch_len(pot24 - 1, pot24, "pot24 - 1");
             ensure_sketch_len(pot24, pot24, "pot24");
             ensure_sketch_len(pot(27), pot24, "pot(27)");
-            ensure_sketch_len(usize::MAX, pot24, "usize::MAX");
+            ensure_sketch_len(u32::MAX as u64, pot24, "u32::MAX");
         } else {
             // target_pointer_width: 64 or larger.
             let pot30 = pot(30);
@@ -816,7 +1203,7 @@ mod tests {
             // due to ceiling to next_power_of_two
             ensure_sketch_len(pot30 - 1, pot30, "pot30- 1");
             ensure_sketch_len(pot30, pot30, "pot30");
-            ensure_sketch_len(usize::MAX, pot30, "usize::MAX");
+            ensure_sketch_len(u64::MAX, pot30, "u64::MAX");
         };
     }
 }
